/* cfg - persistent configuration properties interface
 * Copyright (c) 2002 Michael B. Allen <mballen@erols.com>
 *
 * The MIT License
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <stdarg.h>
#include <limits.h>
#include "mba/msgno.h"
#include "mba/iterator.h"
#include "mba/linkedlist.h"
#include "mba/cfg.h"
#include "mba/text.h"
#include "defines.h"

#ifdef USE_WCHAR
#define T1PMNF(m,s) PMNF((m), ": %ls", (s))
#else
#define T1PMNF(m,s) PMNF((m), ": %s", (s))
#endif

struct cfg {
	struct linkedlist *list;
	tchar buf[BUFSIZ];
};

struct cfg *
cfg_new(void)
{
	struct cfg *this;
	if ((this = malloc(sizeof *this)) == NULL ||
			(this->list = linkedlist_new(0)) == NULL) {
		PMNO(errno);
		return NULL;
	}
	return this;
}
void
cfg_del(void *this)
{
	if (this) {
		struct cfg *cfg = this;
		linkedlist_del(cfg->list, free);
		free(cfg);
	}
}

static int
readline(tchar *buf, FILE *in)
{
	int ch, idx = 0;

	if ((ch = fgetc(in)) == EOF) {
		return ferror(in) ? -1 : 0;
	}
	if (ch == _T('\n')) {
		buf[0] = _T('\0');
		return 1;
	}
	buf[idx++] = ch;
	while (idx < BUFSIZ && (ch = fgetc(in)) != _T('\n') && ch != -1) {
		buf[idx++] = ch;
	}
	if (idx >= BUFSIZ) {
		errno = E2BIG;
		PMNO(errno);
		return -1;
	}
	buf[idx++] = _T('\0');
	return idx;
}
static int
writeline(tchar *buf, FILE *out)
{
#ifdef USE_WCHAR
	char mb[16];
	int n = 0;

#if __USE_UNIX98
	mbstate_t ps;
    memset(&ps, 0, sizeof(ps));
    while (*buf) {
		if ((n = wcrtomb(mb, *buf, &ps)) != -1) {
#else
    while (*buf) {
		if ((n = wctomb(mb, *buf)) != -1) {
#endif
			PMNO(errno);
			return -1;
		}
		if (fwrite(mb, n, 1, out) != 1) {
			PMNO(errno);
			return -1;
		}
		buf++;
	}
#else
	if (fputs(buf, out) == EOF && ferror(out)) {
		PMNO(errno);
		return -1;
	}
#endif
	return 0;
}
static int
validateline(const tchar *str, const tchar *end)
{
	int state;

	state = 0;
	for ( ; str < end; str++) {
		switch (state) {
			case 0:
				if (*str == _T('\0')) {
					return 1;
				} else if (*str == _T('#') || *str == _T('!')) {
					state = 3;
				} else if (istspace(*str) == 0) {
					state = 1;
				}
				break;
			case 1:
			case 2:
				if (*str == _T('\0')) {
					errno = EINVAL;
					PMNO(errno);
					return -1;
				} else if (*str == _T('=')) {
					state = 3;
				} else if (istspace(*str)) {
					state = 2;
				} else if (state == 2) {
					errno = EINVAL;
					PMNO(errno);
					return -1;
				}
				break;
			case 3:
				if (*str == _T('\0')) {
					return 0;
				}
				break;
		}
	}
	errno = E2BIG;
	PMNO(errno);
	return -1;
}

int
cfg_load_str(struct cfg *this, const tchar *src, const tchar *slim)
{
	const tchar *end;
	tchar *line;
	int row, n;

	if (this == NULL || src == NULL || slim == NULL) {
		errno = EINVAL;
		PMNF(errno, ": this=%p", this);
		return -1;
	}

	end = src;
	for (row = 1; *end; row++) {
		while (*end && *end != _T('\n')) {
			end++;
		}
		if ((n = text_copy_new(src, slim, &line, end - src)) == -1) {
			PMNO(errno);
			return -1;
		}
		if (validateline(line, line + n + 1) == -1 ||
				linkedlist_add(this->list, line) == -1) {
			AMSG("line %d", row);
			linkedlist_clear(this->list, free);
			return -1;
		}
		if (*end == _T('\0')) {
			break;
		}
		end = src = end + 1;
	}

	return 0;
}
#ifndef _WIN32
extern char **environ;
#endif
int
cfg_load_env(struct cfg *this)
{
	char **e;

	if (this == NULL) {
		errno = EINVAL;
		PMNO(errno);
		return -1;
	}

	for (e = environ; *e; e++) {
		tchar *str;

#ifdef USE_WCHAR
		size_t n;

#if __USE_UNIX98
		mbstate_t s;
		const char *var;

		var = *e;
		memset(&s, 0, sizeof(s));
		if ((n = mbsrtowcs(NULL, &var, 0, &s)) == (size_t)-1) {
			PMNO(errno);
			return -1;
		}
		n++;

		str = malloc(n * sizeof *str);
		memset(&s, 0, sizeof(s));
		if ((n = mbsrtowcs(str, &var, n, &s)) == (size_t)-1) {
			PMNO(errno);
			free(str);
			return -1;
		}
#else
		if ((n = mbstowcs(NULL, *e, 0)) == (size_t)-1) {
			PMNO(errno);
			return -1;
		}
		n++;

		str = malloc(n * sizeof *str);
		if ((n = mbstowcs(str, *e, n)) == (size_t)-1) {
			PMNO(errno);
			free(str);
			return -1;
		}
#endif
#else
		if ((str = strdup(*e)) == NULL) {
			PMNO(errno);
			return -1;
		}
#endif

		if (validateline(str, str + BUFSIZ) == -1 ||
				linkedlist_add(this->list, str) == -1) {
			AMSG("%s", *e);
			linkedlist_clear(this->list, free);
			free(str);
			return -1;
		}
	}

	return 0;
}
int
cfg_load_cgi_query_string(struct cfg *this, const tchar *qs)
{
	const tchar *s;
	int state, bi;
	tchar buf[8192];

	if (this == NULL || qs == NULL) {
		errno = EINVAL;
		PMNO(errno);
		return -1;
	}

	state = 0;
	bi = 0;
	s = qs;
	do {
		buf[bi] = *qs;
		switch (state) {
			case 0:
				if (*qs == _T('\0')) {
					return 0;
				}
				if (*qs == _T('&') || *qs == _T('=') || istprint(*qs) == 0) {
					errno = EINVAL;
					T1PMNF(errno, qs);
					return -1;
				}
				state = 1;
				break;
			case 1:
				if (*qs == _T('=')) {
					state = 2;
					break;
				} else if (*qs == _T('&')) {
					errno = EINVAL;
					T1PMNF(errno, qs);
					return -1;
				}
				break;
			case 2:
				if (*qs == _T('&') || *qs == _T('\0')) {
					buf[bi] = _T('\0');
					if (validateline(buf, buf + 8192) == -1 ||
							linkedlist_add(this->list, tcsdup(buf)) == -1) {
#ifdef USE_WCHAR
						AMSG("%ls", buf);
#else
						AMSG("%s", buf);
#endif
						return -1;
					}
					if (*qs == _T('\0')) {
						return 0;
					}
					bi = 0;
					state = 0;
					break;
				} else if (*qs == _T('=')) {
					errno = EINVAL;
					T1PMNF(errno, qs);
					return -1;
				}
				break;
		}
		if (state != 0) {
			bi++;
		}
	} while (*qs++);
	return 0;
}
int
cfg_load(struct cfg *this, const char *filename)
{
	FILE *fp;
	int row, n;
	tchar buf[BUFSIZ];

	if (this == NULL || filename == NULL) {
		errno = EINVAL;
		PMNF(errno, ": this=%p", this);
		return -1;
	}

	fp = fopen(filename, "r");
	if (fp == NULL) {
		PMNF(errno, ": %s", filename);
		return -1;
	}

	for (row = 1;; row++) {
		if ((n = readline(buf, fp)) == -1) {
			AMSG("");
			fclose(fp);
			return -1;
		} else if (n == 0) {
			break;
		}
		if (validateline(buf, buf + BUFSIZ) == -1 ||
					linkedlist_add(this->list, tcsdup(buf)) == -1) {
			AMSG("%s: line %d", filename, row);
			linkedlist_clear(this->list, free);
			fclose(fp);
			return -1;
		}
	}
	fclose(fp);
	return 0;
}
int
cfg_write(struct cfg *this, FILE *stream)
{
	tchar *line;
	iter_t iter;
	int n;

	if (this == NULL || stream == NULL) {
		errno = EINVAL;
		PMNO(errno);
		return -1;
	}

	linkedlist_iterate(this->list, &iter);
	while ((line = linkedlist_next(this->list, &iter)) != NULL) {
		if ((n = writeline(line, stream)) == -1) {
			AMSG("");
			return -1;
		}
		fputc('\n', stream);
	}

	return 0;
}
int
cfg_store(struct cfg *this, const char *filename)
{
	FILE *fp;
	int ret;

	if (this == NULL || filename == NULL) {
		errno = EINVAL;
		PMNF(errno, ": this=%p", this);
		return -1;
	}

	fp = fopen(filename, "w");
	if (fp == NULL) {
		PMNF(errno, ": %s", filename);
		return -1;
	}

	ret = cfg_write(this, fp);

	fclose(fp);

	return ret;
}

int
cfg_get_str(struct cfg *this, tchar *dst, int dn, const tchar *def, const tchar *name)
{
	tchar *str;
	const tchar *p, *end;
	int state;
	iter_t iter;

	if (this == NULL || dst == NULL || name == NULL || *name == _T('\0')) {
		errno = EINVAL;
		PMNO(errno);
		return -1;
	}

	linkedlist_iterate(this->list, &iter);
	while ((str = linkedlist_next(this->list, &iter)) != NULL) {
		state = 0;
		p = name;
		for ( ; state != 5; str++) {
			switch (state) {
				case 0:
					if (*str == _T('\0') || *str == _T('!') || *str == _T('#')) {
						state = 5;
						break;
					}
					if (istspace(*str)) {
						break;
					}
					state = 1;
				case 1:
					if (*p == _T('\0') && (istspace(*str) || *str == _T('='))) {
						state = 2;
					} else if (*str == *p) {
						p++;
						break;
					} else {
						state = 5;
						break;
					}
				case 2:
					if (*str == _T('=')) {
						state = 3;
					}
					break;
				case 3:
					if (istspace(*str)) {
						break;
					}
				case 4:
					for (p = end = str; *p != _T('\0'); p++) {
						if (istspace(*p) == 0) {
							end = p;
						}
					}
					end++;
					if ((end - str) >= dn) {
						errno = ERANGE;
						PMNO(errno);
						return -1;
					}
					dn = end - str;
					memcpy(dst, str, dn * sizeof *dst);
					dst[dn] = _T('\0');
					return 0;
			}
		}
	}
	if (def) {
		tcsncpy(dst, def, dn);
	} else {
		errno = EFAULT;
		T1PMNF(errno, name);
		return -1;
	}
	return 0;
}
int
cfg_vget_str(struct cfg *this, tchar *dst, int dn,
			const tchar *def, const tchar *name, ...)
{
	tchar buf[128];
	va_list ap;

	va_start(ap, name);

	if (vstprintf(buf, 128, name, ap) == -1) {
		PMNO(errno);
		return -1;
	}
	if (cfg_get_str(this, dst, dn, def, buf) == -1) {
		AMSG("");
		return -1;
	}

	va_end(ap);
	return 0;
}
int
cfg_get_short(struct cfg *this, short *dst, short def, const tchar *name)
{
	long ul;

	if (cfg_get_long(this, &ul, def, name) == -1) {
		AMSG("");
		return -1;
	}
	*dst = ul & 0xFFFF;
	return 0;
}
int
cfg_get_int(struct cfg *this, int *dst, int def, const tchar *name)
{
	long ul;

	if (cfg_get_long(this, &ul, def, name) == -1) {
		AMSG("");
		return -1;
	}
	*dst = ul;
	return 0;
}
int
cfg_get_long(struct cfg *this, long *dst, long def, const tchar *name)
{
	long ret;
	tchar buf[16];

	if (cfg_get_str(this, buf, 16, NULL, name) == 0) {
		if ((ret = tcstol(buf, NULL, 0)) == LONG_MIN || ret == LONG_MAX) {
			PMNO(errno);
			return -1;
		}
		*dst = ret;
	} else {
		*dst = def;
	}
	return 0;
}
int
cfg_vget_short(struct cfg *this, short *dst, short def, const tchar *name, ...)
{
	tchar buf[128];
	va_list ap;

	va_start(ap, name);

	if (vstprintf(buf, 128, name, ap) == -1) {
		PMNO(errno);
		return -1;
	}
	if (cfg_get_short(this, dst, def, buf) == -1) {
		AMSG("");
		return -1;
	}

	va_end(ap);
	return 0;
}
int
cfg_vget_int(struct cfg *this, int *dst, int def, const tchar *name, ...)
{
	tchar buf[128];
	va_list ap;

	va_start(ap, name);

	if (vstprintf(buf, 128, name, ap) == -1) {
		PMNO(errno);
		return -1;
	}
	if (cfg_get_int(this, dst, def, buf) == -1) {
		AMSG("");
		return -1;
	}

	va_end(ap);
	return 0;
}
int
cfg_vget_long(struct cfg *this, long *dst, long def, const tchar *name, ...)
{
	tchar buf[128];
	va_list ap;

	va_start(ap, name);

	if (vstprintf(buf, 128, name, ap) == -1) {
		PMNO(errno);
		return -1;
	}
	if (cfg_get_long(this, dst, def, buf) == -1) {
		AMSG("");
		return -1;
	}

	va_end(ap);
	return 0;
}
void
cfg_iterate(void *cfg, iter_t *iter)
{
	struct cfg *this = cfg;
	if (this) {
		linkedlist_iterate(this->list, iter);
	}
}
const tchar *
cfg_next(void *cfg, iter_t *iter)
{
	struct cfg *this = cfg;
	tchar *str, *p;
	int state;

	if (this == NULL) {
		return NULL;
	}

	while ((str = linkedlist_next(this->list, iter)) != NULL) {
		state = 0;
		p = this->buf;
		for ( ; state != 2; str++) {
			switch (state) {
				case 0:
					if (*str == _T('\0') || *str == _T('!') || *str == _T('#')) {
						state = 2;
						break;
					}
					if (istspace(*str)) {
						break;
					}
					state = 1;
				case 1:
					if ((istspace(*str) || *str == _T('='))) {
						*p = _T('\0');
						return this->buf;
					} else if ((p - this->buf) == BUFSIZ) {
						return NULL;
					}
					*p++ = *str;
			}
		}
	}

	return NULL;
}

