/* varray - a variable sized array
 * Copyright (c) 2003 Michael B. Allen <mballen@erols.com>
 *
 * The MIT License
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "mba/iterator.h"
#include "mba/varray.h"
#include "defines.h"

struct varray {
	size_t size;                                          /* element size */
	void *bins[16];                                 /* 0 to 2^21 elements */
};

struct varray *
varray_new(size_t membsize)
{
	struct varray *va;

	if (membsize == 0) {
		errno = EINVAL;
		return NULL;
	}
	if ((va = calloc(sizeof *va, 1)) == NULL) {
		return NULL;
	}
	va->size = membsize;

	return va;
}
void
varray_del(void *va)
{
	varray_release(va, 0);
	free(va);
}
void
varray_release(struct varray *va, unsigned int from)
{
	unsigned int r, i;

	if (va == NULL) {
		return;
	}

	r = 32;
	for (i = 0; i < 16; i++) {
		if (from <= r) {
			break;
		}
		r *= 2;
	}
	if (from != 0 ) i++;
	for ( ; i < 16; i++) {
		if (va->bins[i]) {
			free(va->bins[i]);
			va->bins[i] = NULL;
		}
	}
}
#include "mba/msgno.h"
void *
varray_get(struct varray *va, unsigned int idx)
{
	unsigned int r, i, n;

	if (va == NULL) {
		errno = EINVAL;
		return NULL;
	}

	r = 32;          /* First and second bins hold 32 then 64,128,256,... */
	for (i = 0; i < 16; i++) {
		if (r > idx) {
			break;
		}
		r *= 2;
	}
	if (i == 16) {
		errno = ERANGE;
		return NULL;
	}

	n = i == 0 ? 32 : 1 << (i + 4);                 /* n is nmemb in bin i */

	if (va->bins[i] == NULL && (va->bins[i] = calloc(n, va->size)) == NULL) {
		errno = ENOMEM;
		return NULL;
	}

	return (char *)va->bins[i] + (i == 0 ? idx : idx - n) * va->size;
}
void
varray_iterate(void *this, iter_t *iter)
{
	this = NULL;
	iter->i1 = iter->i2 = 0;
}
void *
varray_next(void *this, iter_t *iter)
{
	struct varray *va = this;
	unsigned int n;

	n = iter->i1 == 0 ? 32 : 1 << (iter->i1 + 4);           /* n is nmemb in iter->i1 */

	if (iter->i2 == n) {
		iter->i2 = 0;
		iter->i1++;
	}

	while (va->bins[iter->i1] == NULL) {
		iter->i1++;
		if (iter->i1 == 16) {
			return NULL;
		}
	}

	return (char *)va->bins[iter->i1] + iter->i2++ * va->size;
}

